#ifndef INC_ERA_MQTT_LINUX_HPP_
#define INC_ERA_MQTT_LINUX_HPP_

#include <Utility/ERaUtility.hpp>
#include <Utility/ERacJSON.hpp>
#include <ERa/ERaDefine.hpp>
#include <ERa/ERaConfig.hpp>
#include <ERa/ERaApi.hpp>
#include "MQTT/MQTT.hpp"
#include "ERaMqttHelper.hpp"

#if defined(__has_include) &&       \
    __has_include(<functional>) &&  \
    !defined(ERA_IGNORE_STD_FUNCTIONAL_STRING)
    #include <functional>
    #define MQTT_HAS_FUNCTIONAL_H
#endif

#define ERA_MQTT_PUB_LOG(status, errorCode)                                                                                     \
    if (status) {                                                                                                               \
        ERA_LOG(TAG, ERA_PSTR("Publish (ok #%d) %s: %s"), this->mqtt.lastPacketID(), topic, payload);                           \
    }                                                                                                                           \
    else {                                                                                                                      \
        ERA_LOG_ERROR(TAG, ERA_PSTR("Publish (error %d #%d) %s: %s"), errorCode, this->mqtt.lastPacketID(), topic, payload);    \
    }

#define ERA_MQTT_SUB_LOG(status, errorCode)                                                                                     \
    if (status) {                                                                                                               \
        ERA_LOG(TAG, ERA_PSTR("Subscribe (ok): %s, QoS: %d"), topicName, qos);                                                  \
    }                                                                                                                           \
    else {                                                                                                                      \
        ERA_LOG_ERROR(TAG, ERA_PSTR("Subscribe (error %d): %s, QoS: %d"), errorCode, topicName, qos);                           \
    }

#define ERA_MQTT_UNSUB_LOG(status, errorCode)                                                                                   \
    if (status) {                                                                                                               \
        ERA_LOG(TAG, ERA_PSTR("Unsubscribe (ok): %s"), topicName);                                                              \
    }                                                                                                                           \
    else {                                                                                                                      \
        ERA_LOG_ERROR(TAG, ERA_PSTR("Unsubscribe (error %d): %s"), errorCode, topicName);                                       \
    }

using namespace std;

template <class MQTT>
class ERaMqttLinux
    : public ERaMqttHelper
{
#if defined(MQTT_HAS_FUNCTIONAL_H)
    typedef std::function<void(void)> StateCallback_t;
    typedef std::function<void(const char*, const char*)> MessageCallback_t;
#else
    typedef void (*StateCallback_t)(void);
    typedef void (*MessageCallback_t)(const char*, const char*);
#endif

    const char* TAG = "MQTT";
    const char* ONLINE_MESSAGE = R"json({"ol":1%s%s})json";
    const char* OFFLINE_MESSAGE = R"json({"ol":0})json";
    const char* WIFI_INFO = R"json(,"wifi_ssid":"%s")json";
    const char* ASK_CONFIG_INFO = R"json(,"ask_configuration":1)json";
    const char* LWT_TOPIC = "/is_online";
    const bool LWT_RETAINED = true;
    const int LWT_QOS = QoST::QOS1;

public:
    ERaMqttLinux()
        : mqtt(0)
        , host(ERA_MQTT_HOST)
        , port(ERA_MQTT_PORT)
        , clientID(ERA_MQTT_CLIENT_ID)
        , username(ERA_MQTT_USERNAME)
        , password(ERA_MQTT_PASSWORD)
        , ERaTopic(ERA_MQTT_BASE_TOPIC)
        , ERaAuth(ERA_AUTHENTICATION_TOKEN)
        , ssid(NULL)
        , ping(0L)
        , lastPublish(0L)
        , signalQuality(0)
        , askConfig(false)
        , needPubState(false)
        , _connected(false)
        , mutex(NULL)
        , connectedCb(NULL)
        , disconnectedCb(NULL)
    {
        this->mqtt.setKeepAlive(ERA_MQTT_KEEP_ALIVE);
        memset(this->willTopic, 0, sizeof(this->willTopic));
    }
    ~ERaMqttLinux()
    {}

    void begin(const char* _host, uint16_t _port,
            const char* _username, const char* _password) {
        this->config(_host, _port, _username, _password);
    }

    void init(int readBufSize = ERA_MQTT_RX_BUFFER_SIZE,
            int writeBufSize = ERA_MQTT_TX_BUFFER_SIZE) {
        this->mqtt.init(readBufSize, writeBufSize);
    }

    void config(const char* _host, uint16_t _port,
            const char* _username, const char* _password);
    bool connect();
    void disconnect();
    bool connected();
    bool run();
    bool subscribeTopic(const char* baseTopic, const char* topic,
                        QoST qos = (QoST)ERA_MQTT_SUBSCRIBE_QOS) override;
    bool unsubscribeTopic(const char* baseTopic, const char* topic) override;
    bool publishData(const char* topic, const char* payload,
                    bool retained = ERA_MQTT_PUBLISH_RETAINED,
                    QoST qos = (QoST)ERA_MQTT_PUBLISH_QOS) override;
    bool publishState(bool online);
    bool syncConfig();

    void setTimeout(uint32_t timeout) {
        this->mqtt.setTimeout(timeout);
    }

    void setSkipACK(bool skip = true) {
        this->mqtt.setSkipACK(skip);
    }

    void setKeepAlive(uint16_t keepAlive) {
        this->mqtt.setKeepAlive(keepAlive);
    }

    void setClientID(const char* id) {
        this->clientID = id;
    }

    const char* getClientID() const {
        if (this->clientID == NULL) {
            return "";
        }
        return this->clientID;
    }

    void setTopic(const char* topic) {
        this->ERaTopic = topic;
    }

    void setAuth(const char* auth) {
        this->ERaAuth = auth;
    }

    const char* getAuth() const {
        return this->ERaAuth;
    }

    void setSSID(const char* _ssid) {
        this->ssid = _ssid;
    }

    const char* getSSID() const {
        return this->ssid;
    }

    MillisTime_t getPing() const {
        return this->ping;
    }

    void setSignalQuality(int16_t signal) {
        this->signalQuality = signal;
    }

    int16_t getSignalQuality() const {
        return this->signalQuality;
    }

    void setAskConfig(bool enable) {
        this->askConfig = enable;
    }

    bool getAskConfig() const {
        return this->askConfig;
    }

    const char* getTag() const {
        return this->TAG;
    }

    const char* getLWTTopic() const {
        return this->LWT_TOPIC;
    }

    const char* getLWTPayload() const {
        return this->OFFLINE_MESSAGE;
    }

    void onMessage(MessageCallback_t cb) {
        this->mqtt.onMessage(cb);
    }

    void onStateChange(StateCallback_t onCb,
                       StateCallback_t offCb) {
        this->connectedCb = onCb;
        this->disconnectedCb = offCb;
    }

    bool connectionDenied() {
        return ((this->mqtt.lastError() == lwmqtt_err_t::LWMQTT_MISSING_OR_WRONG_PACKET) ||
                (this->mqtt.lastError() == lwmqtt_err_t::LWMQTT_CONNECTION_DENIED));
    }

protected:
private:
    bool publishLWT(bool sync = false);
    void onConnected();
    void onDisconnected();
    void lockMQTT();
    void unlockMQTT();

    MQTT mqtt;
    const char* host;
    uint16_t port;
    const char* clientID;
    const char* username;
    const char* password;
    const char* ERaTopic;
    const char* ERaAuth;
    const char* ssid;
    MillisTime_t ping;
    MillisTime_t lastPublish;
    int16_t signalQuality;
    bool askConfig;
    bool needPubState;
    bool _connected;
    char willTopic[MAX_TOPIC_LENGTH];
    ERaMutex_t mutex;
    StateCallback_t connectedCb;
    StateCallback_t disconnectedCb;
};

template <class MQTT>
inline
void ERaMqttLinux<MQTT>::config(const char* _host, uint16_t _port,
                            const char* _username, const char* _password) {
    this->host = _host;
    this->port = _port;
    this->username = _username;
    this->password = _password;
    ClearArray(this->willTopic);
    FormatString(this->willTopic, this->ERaTopic);
    FormatString(this->willTopic, LWT_TOPIC);
    this->mqtt.init(ERA_MQTT_RX_BUFFER_SIZE,
                    ERA_MQTT_TX_BUFFER_SIZE);
    this->mqtt.setWill(this->willTopic, OFFLINE_MESSAGE, LWT_RETAINED, LWT_QOS);
    this->mqtt.begin(this->host, this->port);
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::connect() {
    size_t count {0};
    char _clientID[74] {0};
    if (this->clientID != NULL) {
        FormatString(_clientID, this->ERaAuth);
        if (!ERaStrCmp(this->clientID, this->ERaAuth) &&
            strlen(this->clientID)) {
            FormatString(_clientID, "_%s", this->clientID);
        }
    }

    this->mqtt.disconnect();
    this->_connected = false;

    while (this->mqtt.connect(_clientID, this->username, this->password) == false) {
        ERA_LOG_ERROR(TAG, ERA_PSTR("MQTT(%d): Connect failed (%d), retrying in 5 seconds"),
                                    (count + 1), this->mqtt.lastError());
        ERaDelay(5000);
        if (++count >= LIMIT_CONNECT_BROKER_MQTT) {
            return false;
        }
    }

#if defined(ERA_ZIGBEE) ||  \
    defined(ERA_SPECIFIC)
    subscribeTopic(this->ERaTopic, "/zigbee/+/get");
    subscribeTopic(this->ERaTopic, "/zigbee/+/down");
    subscribeTopic(this->ERaTopic, "/zigbee/permit_to_join");
    subscribeTopic(this->ERaTopic, "/zigbee/remove_device");
#endif

    subscribeTopic(this->ERaTopic, "/arduino_pin/+");
    subscribeTopic(this->ERaTopic, "/virtual_pin/+");
    subscribeTopic(this->ERaTopic, "/+/down");
    subscribeTopic(this->ERaTopic, "/down");

#if defined(ERA_ASK_CONFIG_WHEN_RESTART)
    if (!this->publishLWT(true)) {
        return false;
    }
#else
    if (!this->publishLWT(this->getAskConfig())) {
        return false;
    }
#endif

    subscribeTopic(this->ERaTopic, LWT_TOPIC);
    ERA_LOG(TAG, ERA_PSTR(R"(Connected: "%s:%d", Client ID: "%s", Username: "%s", Password: "%s")"),
                            this->host, this->port, _clientID, this->username, this->password);

    this->onConnected();
    return true;
}

template <class MQTT>
inline
void ERaMqttLinux<MQTT>::disconnect() {
    this->mqtt.disconnect();
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::connected() {
    return this->mqtt.connected();
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::run() {
    if (!this->_connected) {
        return false;
    }

    if (!this->mqtt.loop()) {
        this->onDisconnected();
        return this->connect();
    }
    else if (this->needPubState) {
        this->publishLWT(false);
        this->needPubState = false;
    }
    return true;
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::subscribeTopic(const char* baseTopic, const char* topic,
                                        QoST qos) {
    bool status {false};
    char topicName[MAX_TOPIC_LENGTH] {0};
    if (baseTopic != NULL) {
	    FormatString(topicName, baseTopic);
    }
    if (topic != NULL) {
	    FormatString(topicName, topic);
    }

    this->lockMQTT();
    if (this->connected()) {
        status = this->mqtt.subscribe(topicName, qos);
        ERA_MQTT_SUB_LOG(status, this->mqtt.lastError())
    }
    this->unlockMQTT();

    return status;
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::unsubscribeTopic(const char* baseTopic, const char* topic) {
    bool status {false};
    char topicName[MAX_TOPIC_LENGTH] {0};
    if (baseTopic != NULL) {
	    FormatString(topicName, baseTopic);
    }
    if (topic != NULL) {
	    FormatString(topicName, topic);
    }

    this->lockMQTT();
    if (this->connected()) {
        status = this->mqtt.unsubscribe(topicName);
        ERA_MQTT_UNSUB_LOG(status, this->mqtt.lastError())
    }
    this->unlockMQTT();

    return status;
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::publishData(const char* topic, const char* payload,
                                    bool retained, QoST qos) {
    if (!this->_connected) {
        return false;
    }

    bool status {false};

    this->lockMQTT();
    if (this->connected()) {
        this->lastPublish = ERaMillis();
        status = this->mqtt.publish(topic, payload, retained, qos);
        this->ping = (ERaMillis() - this->lastPublish);
        ERA_MQTT_PUB_LOG(status, this->mqtt.lastError())
    }
    this->unlockMQTT();

    return status;
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::publishState(bool online) {
    this->needPubState = online;
    return true;
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::syncConfig() {
    return this->publishLWT(true);
}

template <class MQTT>
inline
bool ERaMqttLinux<MQTT>::publishLWT(bool sync) {
    bool status {false};
    char wifiInfo[50] {0};
    char payload[100] {0};
    char* topic = this->willTopic;

    if (this->getSSID() != NULL) {
        FormatString(wifiInfo, WIFI_INFO, this->getSSID());
    }

    FormatString(payload, ONLINE_MESSAGE, wifiInfo, sync ? ASK_CONFIG_INFO : "");

    this->lockMQTT();
    if (this->connected()) {
        this->lastPublish = ERaMillis();
        status = this->mqtt.publish(topic, payload, LWT_RETAINED, LWT_QOS);
        this->ping = (ERaMillis() - this->lastPublish);
        ERA_MQTT_PUB_LOG(status, this->mqtt.lastError())
    }
    this->unlockMQTT();

    return status;
}

template <class MQTT>
inline
void ERaMqttLinux<MQTT>::onConnected() {
    if (this->_connected) {
        return;
    }
    this->_connected = true;
    if (this->connectedCb == NULL) {
        return;
    }
    this->connectedCb();
}

template <class MQTT>
inline
void ERaMqttLinux<MQTT>::onDisconnected() {
    if (!this->_connected) {
        return;
    }
    this->_connected = false;
    if (this->disconnectedCb == NULL) {
        return;
    }
    this->disconnectedCb();
}

template <class MQTT>
inline
void ERaMqttLinux<MQTT>::lockMQTT() {
    if (!this->mqtt.getSkipACK()) {
        return;
    }
    ERaGuardLock(this->mutex);
}

template <class MQTT>
inline
void ERaMqttLinux<MQTT>::unlockMQTT() {
    if (!this->mqtt.getSkipACK()) {
        return;
    }
    ERaGuardUnlock(this->mutex);
}

#define ERaMqtt ERaMqttLinux

#endif /* INC_ERA_MQTT_LINUX_HPP_ */
